#include <windows.h>
#include <windef.h>
#include <stdio.h>
#include <malloc.h>
#include <commctrl.h>

#include "trayem.h"

//#include "win32process.h"

#include "resource.h"

HWND hWndMe = NULL;
HWND hWndParent = NULL;
HINSTANCE hModule = NULL;
TCHAR szTitle[] = "TrayTasks";
TCHAR szWindowClass[] = "TrayTasksContainer";
TCHAR szNotifWindowClass[] = "TrayTasksNotifiedWindow";

const char rcsRevision[] = "$Revision: 1.8 $"; // Our Version 
const char rcsId[] = "$Id: TrayTasks 1.8 2004/04/05 (Sergey Trofimov, Seg@) $"; // The Full RCS ID.

#define MAX_WINPOS 3000

//Win32Process m_win32proc;

// Configurable vars
HICON hDefIcon = NULL;        // Icon for apps without icons
HICON hSepIcon = NULL;        // Separator icon

HICON hEmptyIcon = NULL;

HICON hCurSelIcon = NULL;     // Selected icon
HICON hCurSelIconOriginal = NULL;

BOOL  bModuleEnabled = TRUE;
BOOL  bModuleInitialized = FALSE;
BOOL  bAutoReorder = TRUE;    // Auto reordering
BOOL  bHighlight = TRUE;      // Foreground task highlighting
BOOL  bSwitchAlways = FALSE;  // Enable switching in VWM
BOOL  bLargeIcons = FALSE;    // Using large app icons first 
BOOL  bSeparate    = FALSE;   // Separating enabler
int   nSepCount    = 0;       // Number of sep icons.
                              // Dynamic in tray room mode
int   nStaticSepCount = 1;
BOOL  bTrayedRoom  = FALSE;   // tray room mode
int   nTrayedRoomSize  = 0;   //
BOOL  bFixedTray = FALSE;     // fixed size tray mode
int   nFixedTraySize = 0;
UINT  nTrayIconsCount = 0;     // Number of all icons in systray
                               // Set in FindLastTrayWindow()
BOOL  bFilterCaptions  = FALSE;
BOOL  bFilterClasses   = FALSE;
BOOL  bFilterProcesses = FALSE;
BOOL  bAutoMinimize    = TRUE;
BOOL  bHideMinAppBar   = FALSE;
BOOL  bWaitingForReorder = FALSE;
BOOL  bWaitingForRecalc = FALSE;

BOOL  bUseFlashing;
BOOL  bFlashAsSelected;


TaskListType        taskList;       // Trayed tasks
SpecialWndListType  shownWndList,   // Always shown windows
                    hiddenWndList;  // Always hidden windows

TrayWndListType     trayWndList;    // Really trayed windows

SMouseStroke        msCloseOn      = {false, false, false, 0};
SMouseStroke        msMinimizeOn   = {false, false, false, 0};
SMouseStroke        msUnMaximizeOn = {false, false, false, 0};

bool bTrayIsReordering = false;
bool bIsIdle  = false;
bool bIsFlashing = false;
//bool bFirstTime = true;


#define CONFIG_STRING_LEN 256
#pragma comment(lib,"lsapi")

UINT Msgs[] =
{
	LM_GETREVID,
	LM_SYSTRAY,
	LM_RESTORESYSTRAY,
	LM_SAVESYSTRAY,
	LM_REDRAW,
	0
};

int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd)
{
	return initModuleEx (ParentWnd, dllInst, wd->lsPath);
}


int initModuleEx (HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	hModule = dllInst;
	hWndParent = ParentWnd;

	AddBangCommand("!TRAYTASKSTOGGLE", BangTrayTasksToggle);

	hEmptyIcon = LoadIcon(hModule, MAKEINTRESOURCE(IDI_EMPTY));

	ReadConfigData();

	MyRegisterClass(hModule);
	MyRegisterNotifClass(hModule);

	hWndMe = CreateWindow(
		szWindowClass,
		szTitle,
		WS_CHILD,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		hWndParent,
		NULL,
		hModule,
		NULL);

	SetWindowLong(hWndMe, GWL_USERDATA, magicDWord);

	SendMessage(hWndParent, LM_REGISTERMESSAGE, (WPARAM) hWndMe, (LPARAM) Msgs);

	// Pause for loading
	SetTimer(hWndMe, INIT_TIMER, 500, 0);

	/* TODO later
	if(bFilterProcesses)
		m_win32proc.Init();
	*/

	return 0;
}


void quitModule(HINSTANCE dllInst)
{
	FreeConfigData();
	SetModuleOff();
	trayWndList.clear();
	SendMessage(hWndParent, LM_UNREGISTERMESSAGE, (WPARAM) hWndMe, (LPARAM) Msgs);
	DestroyWindow(hWndMe);
	if (hCurSelIcon)
	{
		DestroyIcon(hCurSelIcon);
		hCurSelIcon = NULL;
	}
	UnregisterClass(szNotifWindowClass, hModule);
	UnregisterClass(szWindowClass, hModule);
	RemoveBangCommand("!TRAYTASKSTOGGLE");
}


void ReadConfigData()
{
  char szBuf[CONFIG_STRING_LEN];

  //////////////////////////////////////
  // Configuring TrayTasksDefaultIcon
  //
  GetRCString("TrayTasksDefaultIcon", szBuf, "0", CONFIG_STRING_LEN);
  
  if(0 == strcmp(szBuf, "0"))
    hDefIcon = LoadIcon(NULL, IDI_APPLICATION);
  else
  {
    if(0 == strcmp(szBuf, "1"))
      hDefIcon = LoadIcon(NULL, IDI_WINLOGO);
    else
      hDefIcon = LoadLSIcon(szBuf, NULL);
  }

  if(!hDefIcon)
  {
    hDefIcon = LoadIcon(NULL, IDI_APPLICATION);
  }

  
  //////////////////////////////////////
  // Configuring TrayTasks Separating
  //
  GetRCString("TrayTasksSepIcon", szBuf, "", CONFIG_STRING_LEN);
  hSepIcon = LoadLSIcon(szBuf, NULL);
  if(!hSepIcon)
    hSepIcon = LoadIcon(hModule, MAKEINTRESOURCE(IDI_EMPTY));

  nStaticSepCount = GetRCInt("TrayTasksSepCount", 1);

  bSeparate = GetRCBool("TrayTasksSeparate", TRUE);

  nTrayedRoomSize = GetRCInt("TrayTasksTrayedRoom", 0);
  if(nTrayedRoomSize)
  {
    bTrayedRoom = TRUE;
    nSepCount   = 0;
  }

  if(!bTrayedRoom)
  {
    nFixedTraySize = GetRCInt("TrayTasksSystraySize", 0);
    if(nFixedTraySize)
    {
      bFixedTray = TRUE;
      nSepCount   = 0;
    }
  }

  //////////////////////////////////////
  // Configuring TrayTasksAutoReorder
  //
  bAutoReorder = GetRCBool("TrayTasksNoReorder", FALSE);


  //////////////////////////////////////
  // Configuring TrayTasksNoHighlight
  //
  bHighlight = GetRCBool("TrayTasksNoHighlight", FALSE);


  //////////////////////////////////////
  // Configuring TrayTasksSwitchAlways
  //
  bSwitchAlways = GetRCBool("TrayTasksSwitchAlways", TRUE);


  //////////////////////////////////////
  // Configuring TrayTasksUseFlashing
  //
  bUseFlashing = GetRCBool("TrayTasksUseFlashing", TRUE);
  bFlashAsSelected = GetRCBool("TrayTasksFlashAsSelected", TRUE);

  //////////////////////////////////////
  // Configuring TrayTasksStartHidden
  //
  bModuleEnabled = GetRCBool("TrayTasksStartHidden", FALSE);


  //////////////////////////////////////
  // Configuring TrayTasksLargeIcons
  //
  bLargeIcons = GetRCBool("TrayTasksLargeIcons", TRUE);


  //////////////////////////////////////
  // Configuring TrayTasksNoAutoMinimize
  //
  bAutoMinimize = GetRCBool("TrayTasksNoAutoMinimize", FALSE);


  //////////////////////////////////////
  // Configuring TrayTasksHideMinAppBar
  //
  bHideMinAppBar = GetRCBool("TrayTasksHideMinAppBar", TRUE);


  //////////////////////////////////////
  // Configuring special windows
  //
  FILE *f;
  
  f = LCOpen(NULL);
  if(f)
  {
    char  token1[CONFIG_STRING_LEN], // "*TrayTasks"
          token2[CONFIG_STRING_LEN], // Pattern
          token3[CONFIG_STRING_LEN], // Type (Title, App, Class)
          token4[CONFIG_STRING_LEN]; // Action (Hidden, Shown)
    char*	tokens[4];
    tokens[0] = token1;
    tokens[1] = token2;
    tokens[2] = token3;
    tokens[3] = token4;

    SpecialWndDataType type;

    while (LCReadNextConfig (f, "*TrayTasks", szBuf, CONFIG_STRING_LEN))
    {
      LCTokenize (szBuf, tokens, 4, NULL);
      
      if(0 == strcmpi(token3, "caption"))
      {
        type = swdCaptionType;
        bFilterCaptions = TRUE;
      }
      else if(0 == strcmpi(token3, "process"))
      {
        type = swdProcessType;
        bFilterProcesses = TRUE;
      }
      else if(0 == strcmpi(token3, "class"))
      {
        type = swdClassType;
        bFilterClasses = TRUE;
      }

      if(0 == strcmpi(token4, "hide"))
      {
        SSpecialWndData *pSpec = new SSpecialWndData(type, token2);
        hiddenWndList.insert(hiddenWndList.end(), pSpec);
        continue;
      }

      if(0 == strcmpi(token4, "show"))
      {
        SSpecialWndData *pSpec = new SSpecialWndData(type, token2);
        shownWndList.insert(shownWndList.end(), pSpec);
        continue;
      }
    }
  }
  LCClose(f);

  //////////////////////////////////////
  // Configuring mouse strokes
  //
  char *strokeCommands[] = {
    "TrayTasksCloseOn", 
    "TrayTasksMinimizeOn", 
    "TrayTasksUnMaximizeOn"
  };

  SMouseStroke *pStrokes[] = {
    &msCloseOn, 
    &msMinimizeOn, 
    &msUnMaximizeOn
  };

  for(int count = 0; count < 3; count++)
  {
    GetRCString(strokeCommands[count], szBuf, "", CONFIG_STRING_LEN);
    {
      char  token1[CONFIG_STRING_LEN],
            token2[CONFIG_STRING_LEN],
            token3[CONFIG_STRING_LEN],
            token4[CONFIG_STRING_LEN];
      char*	tokens[4];
      tokens[0] = token1;
      tokens[1] = token2;
      tokens[2] = token3;
      tokens[3] = token4;

      int nTokens = LCTokenize (szBuf, tokens, 4, NULL);
    
      for(int i = 0; i < nTokens; i++)
      {
        if(0 == strcmpi(tokens[i], "Ctrl"))
        {
          pStrokes[count]->isCtrl = true;
          continue;
        }
        if(0 == strcmpi(tokens[i], "Alt"))
        {
          pStrokes[count]->isAlt = true;
          continue;
        }
        if(0 == strcmpi(tokens[i], "Shift"))
        {
          pStrokes[count]->isShift = true;
          continue;
        }
      
        if(0 == strcmp(tokens[i], "l"))
        {
          pStrokes[count]->message = WM_LBUTTONDOWN;
          continue;
        }
        if(0 == strcmp(tokens[i], "L"))
        {
          pStrokes[count]->message = WM_LBUTTONDBLCLK;
          continue;
        }
        if(0 == strcmp(tokens[i], "m"))
        {
          pStrokes[count]->message = WM_MBUTTONDOWN;
          continue;
        }
        if(0 == strcmp(tokens[i], "M"))
        {
          pStrokes[count]->message = WM_MBUTTONDBLCLK;
          continue;
        }
        if(0 == strcmp(tokens[i], "r"))
        {
          pStrokes[count]->message = WM_RBUTTONDOWN;
          continue;
        }
        if(0 == strcmp(tokens[i], "R"))
          pStrokes[count]->message = WM_RBUTTONDBLCLK;
      }
    }
  }
}


void FreeConfigData()
{
	shownWndList.safe_clear();
	hiddenWndList.safe_clear();
}


void SetModuleOn()
{
  nSepCount   = 0;
  //nTrayIconsCount = 0;

  bModuleEnabled = TRUE;
  SetTimer(hWndMe, MAIN_TIMER, 200, NULL);
  if (bUseFlashing)
	  SetTimer(hWndMe, FLASH_TIMER, 500, NULL);
  PostMessage(hWndMe, LM_RESTORESYSTRAY, 0, 0);
}


void SetModuleOff()
{
	if(!bModuleEnabled) return;
	bModuleEnabled = FALSE;
	KillTimer(hWndMe, MAIN_TIMER);
	KillTimer(hWndMe, FLASH_TIMER);
	taskList.safe_clear();
}


void EnableModule(BOOL b)
{
  (bModuleEnabled) ? SetModuleOff() : SetModuleOn();
}


void BangTrayTasksToggle(HWND caller, const char* args)
{
	if (bModuleInitialized) EnableModule(!bModuleEnabled);
	return;
}

BOOL IsTaskWindow(HWND hwnd)
{
	if (!IsWindow(hwnd)) return FALSE;

	int style = GetWindowLong(hwnd, GWL_STYLE);
	int exstyle = GetWindowLong(hwnd, GWL_EXSTYLE);
	char szStr[100];
	RECT rc;
	GetWindowRect(hwnd, &rc);

	//Include notfy windows - separators hacks
	GetClassName(hwnd, szStr, 100);
	if (0 == strcmpi(szStr, szNotifWindowClass)) return TRUE;

	// Exception for visible windows with 0 size
	// Some programs do so
	//if(rc.left == rc.right || rc.top == rc.bottom)
	//  return FALSE;

	//Exclude invisible windows
	if (!::IsWindowVisible(hwnd)) return FALSE;

	// Exclude child windows which parent shown by TrayTasks
	// This can be happen with filtering
	if (bFilterCaptions || bFilterClasses || bFilterProcesses)
	{
		HWND hWndParent = GetParent(hwnd);
		if(hWndParent && taskList.find_hwnd(hWndParent))
		{
			return FALSE;
		}
	}

	//////////////////////////////
	// Check for special windows
	//
	// 1. by title
	if (bFilterCaptions)
	{
		GetWindowText(hwnd, szStr, 100);  
		if (shownWndList.find_typed_value(swdCaptionType, szStr))
		{
			GetClassName(hwnd, szStr, 100);
			return 0 != strcmp(szStr, "tooltips_class32");
		}
		if (hiddenWndList.find_typed_value(swdCaptionType, szStr)) return FALSE;
	}

	// 2. by class
	if (bFilterClasses)
	{
		GetClassName(hwnd, szStr, 100);
		if (shownWndList.find_typed_value(swdClassType, szStr))
			return TRUE;
		if(hiddenWndList.find_typed_value(swdClassType, szStr))
			return FALSE;
	}

	/* TODO later
	// 3. by process
	if (bFilterProcesses)
	{
		DWORD dwWndPID;
		GetWindowThreadProcessId(hwnd, &dwWndPID);
		m_win32proc.EnumAllProcesses();

		int size = m_win32proc.GetAllProcessesNames()->size();
		for (int i = 0; i < size; i++)
		{
		};

		if (shownWndList.find_typed_value(swdProcessType, szStr))
			return TRUE;
		if (hiddenWndList.find_typed_value(swdProcessType, szStr))
			return FALSE;
	};
	*/

	// Standard determining

	return
        !(style & WS_CHILD) &&
        /*::IsWindowVisible(hwnd) &&*/ 
        (!GetParent(hwnd) || GetParent(hwnd) == GetDesktopWindow()) && 
        GetWindowLong(hwnd, GWL_USERDATA) != magicDWord &&
        !(exstyle & WS_EX_TOOLWINDOW) &&
        //!IsTaskWindow(GetWindow(hwnd, GW_OWNER));
        !GetWindow(hwnd, GW_OWNER);
}


BOOL IsWindowSelected(HWND hwnd)
{
	HWND hWndFore = GetForegroundWindow();
	return hWndFore && (hwnd == hWndFore || hwnd == GetWindow(hWndFore, GW_OWNER));
}


HICON GetIconByLevel(HWND hwnd, BYTE nLevel, int nTimeout)
{
  HICON hIcon = NULL;
  LRESULT res = 0;

  if(!IsWindow(hwnd)) return NULL;

  if(hWndMe == GetParent(hwnd)) return hSepIcon;

  switch(nLevel)
  {
  case 0:
    res = SendMessageTimeout(hwnd, WM_GETICON, (bLargeIcons) ? ICON_BIG : ICON_SMALL, 0, SMTO_NORMAL, nTimeout, (unsigned long *)&hIcon);
    break;

  case 1:
    res = SendMessageTimeout(hwnd, WM_GETICON, (bLargeIcons) ? ICON_SMALL : ICON_BIG, 0, SMTO_NORMAL, nTimeout, (unsigned long *)&hIcon);
    break;

  case 2:
    res = SendMessageTimeout(hwnd, WM_QUERYDRAGICON, 0, 0, SMTO_NORMAL, nTimeout, (unsigned long *)&hIcon);
    break;

  case 3:
    hIcon = (HICON)GetClassLong(hwnd, GCL_HICONSM);
    res = (hIcon) ? 1 : 0;
    break;

  case 4:
    hIcon = (HICON)GetClassLong(hwnd, GCL_HICON);
    res = (hIcon) ? 1 : 0;
    break;
    
  case 5:
    hIcon = hDefIcon;
    res = (hIcon) ? 1 : 0;
    break;
  }

  return (res) ? hIcon : NULL;
}


HICON GetWindowIcon(HWND hwnd, BYTE &nLevel, int nTimeout)
{
	HICON hIcon = NULL;
	LRESULT res = 0;
	int i;
	BYTE nCurLevel = MAXICONLEVEL;

	for (i = 0; i < 6 && i <= nLevel && !hIcon; i++)
	{
		hIcon = GetIconByLevel(hwnd, i, nTimeout);
		nCurLevel = i;
	}

	//if(hIcon)
	//{
		nLevel = nCurLevel;
	//};
	return hIcon;
}


HICON CreateSelectedIcon(HICON hIcon)
{
  if(NULL == hIcon) return NULL;

  ICONINFO iconInfo;
  GetIconInfo(hIcon, &iconInfo);

  BITMAP bm;
  GetObject(iconInfo.hbmColor, sizeof(bm), &bm);

  HIMAGELIST hImg = ImageList_Create(bm.bmWidth, bm.bmHeight, ILC_COLOR | ILC_MASK, 1 ,0);
  ImageList_AddIcon(hImg, hIcon);

  HBITMAP hBmpOld;
  HDC     hMemDC;
  HDC     hScreenDC;

  hScreenDC = GetDC(NULL);
  // Create compatible stuff and select Bitmap
  hMemDC = CreateCompatibleDC(hScreenDC);
  
  hBmpOld = (HBITMAP) SelectObject(hMemDC, iconInfo.hbmColor);
  
  // Draw Image to the compatible DC
  ImageList_Draw(hImg, 0, hMemDC, 0, 0, ILD_SELECTED);
  
  SelectObject(hMemDC, hBmpOld);
  DeleteDC(hMemDC);
  ReleaseDC(NULL, hScreenDC);
  ImageList_Destroy(hImg);

  HICON hIconSel = CreateIconIndirect(&iconInfo);

  DeleteObject(iconInfo.hbmColor);
  DeleteObject(iconInfo.hbmMask);

  return hIconSel;
}


bool IsMyNotifWindow(HWND hwnd)
{
  if (IsWindow(hwnd))
  {
    char szClass[100] = "";
    GetClassName(hwnd, szClass, sizeof(szClass));
    return 0 == strcmp(szClass, szNotifWindowClass);
  }
  return false;
}


BOOL CALLBACK EnumWindowsProc(
	HWND hwnd,      // handle to parent window
	LPARAM lParam   // application-defined value
)
{ 
	if(bHideMinAppBar && IsIconic(hwnd))
	{
		RECT rc;
		GetWindowRect(hwnd, &rc);
		if(rc.left != -MAX_WINPOS || rc.top != MAX_WINPOS)
		{
			SetWindowPos(hwnd, NULL, -MAX_WINPOS, MAX_WINPOS, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE | SWP_ASYNCWINDOWPOS);
		}
	}
  
	if (IsTaskWindow(hwnd) && !taskList.find_hwnd(hwnd))
	{
		// add window to list
		HWND hFrame = (HWND) lParam;
		BYTE nCurIconLevel = MAXICONLEVEL;
  
		HICON hIcon = GetWindowIcon(hwnd, nCurIconLevel, 300);

		if (bHighlight && IsWindowSelected(hwnd))
		{
			if (hCurSelIcon)
			{
				DestroyIcon(hCurSelIcon);
			}
			hCurSelIcon = CreateSelectedIcon(hIcon);
			hCurSelIconOriginal = hIcon;
			hIcon = hCurSelIcon;
		}
  
		if (hIcon)
		{
			HWND hwndNotif = CreateWindow(szNotifWindowClass, NULL, WS_CHILD,
				0, 0, 0, 0, hFrame, NULL, hModule, NULL);
			if(hwndNotif)
			{
				HWND hWndTray=NULL;
				STaskData *pTask  = new STaskData(hwndNotif, hwnd);
				pTask->nIconLevel = nCurIconLevel;
				if (bFixedTray)
				{
					RemoveSep();
				}
				taskList.insert(taskList.end(), pTask);
				pTask->icon.Create(hwndNotif, WM_ICON_NOTIFY, "", hIcon, (UINT) hwnd);
			}
		}
	}
	return TRUE;
}


void OnIdle(HWND hWnd)
{
	TaskListType::iterator i;
	// check for dead trayed apps
	if (bTrayedRoom || bFixedTray)
	{
		TrayWndListType::iterator _L, _F;

		_L = trayWndList.end();
		for(_F = trayWndList.begin(); _F != _L;)
		{
			if (IsWindow((*_F).hwnd))
				++_F;
			else
			{
				UINT nTrayIconsTmp = SendMessage(hWndParent, LM_SENDSYSTRAY, 0, (LPARAM) hWndMe);
				if (nTrayIconsTmp && nTrayIconsTmp < nTrayIconsCount)
				{
					trayWndList.erase(_F++);
					nTrayIconsCount--;
					RefreshState();
				}
				break;
			}
		}
	}
  
	//remove dead
	for(i = taskList.begin(); i != taskList.end(); ++i)
	{
		STaskData *pData = (*i);
		if (!IsTaskWindow(pData->hWnd))
		{
			taskList.safe_erase(i);
			break;
		}
		else if (!pData->m_bFlash && !pData->m_bFlashed && !pData->m_nFlashQueued)
		{
			// refresh tasks' icons
			HICON hIcon = GetWindowIcon(pData->hWnd, pData->nIconLevel, 300);
			if (bHighlight && IsWindowSelected(pData->hWnd))
			{
				if (hIcon != hCurSelIconOriginal)    // foreground changed
				{
					if (hCurSelIcon) DestroyIcon(hCurSelIcon);
					hCurSelIcon = CreateSelectedIcon(hIcon);
				}
				hCurSelIconOriginal = hIcon;
				hIcon = hCurSelIcon;
			}

			if (hIcon && hIcon != pData->icon.GetIcon())
			{
				DestroyIcon(pData->hFlashingIcon);
				pData->hFlashingIcon = CreateSelectedIcon(hIcon);
				pData->icon.SetIcon(hIcon);
			}
		}
	}

	//tuning separators
	if (bSeparate && bAutoReorder)
	{
		BOOL bNeedAddSep = FALSE;
		BOOL bNeedRemSep = FALSE;
		int nAddSepCount = 0;

		if (!bTrayedRoom && !bFixedTray && taskList.size() == 0) // not tray room or fixed size tray mode
		{                                                       // insert seps one time
			bNeedAddSep = TRUE;
			nAddSepCount = nStaticSepCount;
		}
		else if(bTrayedRoom && nTrayedRoomSize)  // tray room mode
		{
			nAddSepCount = - nTrayIconsCount + taskList.size() - nSepCount + nTrayedRoomSize;
		}
		else if(bFixedTray && nFixedTraySize)  // fixed size tray mode
		{
			nAddSepCount = nFixedTraySize - nTrayIconsCount;
		}

		if ((bTrayedRoom || bFixedTray) && (nSepCount + nAddSepCount) >= 0)
		{
			bNeedAddSep = nAddSepCount > 0;
			bNeedRemSep = nAddSepCount < 0;
		}

		if(bNeedRemSep)
		{
			for(int n = 0; n > nAddSepCount; n--)
			{
				nSepCount--;
				TaskListType::iterator b = taskList.begin();
				STaskData *pData = *b;
				if (pData->hWnd == pData->hWndNotif)
				{
					taskList.safe_erase(b);
				}
			}
			//return;
		}
    
		if (bNeedAddSep)
		{
			for(int n = 0; n < nAddSepCount; n++)
			{
				AddSep();
			}
			//taskList.move_to_right();
			ReorderIcons();
			//return;
		}
	}

	//add new
	::EnumWindows(EnumWindowsProc, (LPARAM)hWnd);
}


ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= 0;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= NULL;
	wcex.hCursor		= NULL;
	wcex.hbrBackground	= NULL;
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= NULL;

	return RegisterClassEx(&wcex);
}


ATOM MyRegisterNotifClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= 0;
	wcex.lpfnWndProc	= (WNDPROC)WndProcNotif;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= hSepIcon;
	wcex.hCursor		= NULL;
	wcex.hbrBackground	= NULL;
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= szNotifWindowClass;
	wcex.hIconSm		= NULL;

	return RegisterClassEx(&wcex);
}

void SetFlashing(HWND hWnd, BOOL bFlash)
{
	TaskListType::iterator wi = taskList.begin(), wend = taskList.end();
	for (; wi != wend; wi++)
	{
		if ((*wi)->hWnd == hWnd)
		{
			if (bFlash && (*wi)->m_nFlashQueued<3)
			{
				(*wi)->m_nFlashQueued ++;
			}
			(*wi)->m_bFlash = bFlash != FALSE;
			return;
		}
	}
}

void FlashEmAll()
{
	TaskListType::iterator wi = taskList.begin(), wend = taskList.end();
	for (; wi != wend; wi++)
	{
		if ((*wi)->m_bFlash || (*wi)->m_bFlashed || (*wi)->m_nFlashQueued)
		{
			if ((*wi)->m_nFlashQueued) (*wi)->m_nFlashQueued--;
			(*wi)->m_bFlashed = !((*wi)->m_bFlashed);
			HICON was = (*wi)->icon.GetIcon();
			(*wi)->icon.SetIcon(bFlashAsSelected ? (*wi)->hFlashingIcon : hEmptyIcon);
			if (bFlashAsSelected) (*wi)->hFlashingIcon = was;
		}
	}
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static bool bMainTimerDelayed = false;
	static bool bReorderDelayed = false;
	static bool bRecalcDelayed = false;
	switch (message) 
	{    
		case LM_GETREVID:
		{
			char *buf = (char *) lParam;
			
			if (wParam == 0)
			{
				strcpy(buf, "traytasks.dll (Sergey Trofimov, Seg@): ");
				strcat(buf, &rcsRevision[11]);
				buf[strlen(buf)-1] = '\0';
			}
			else if (wParam == 1)
			{
				strcpy(buf, &rcsId[1]);
				buf[strlen(buf)-1] = '\0';
			} else
			{
				strcpy(buf, "");
			}
			return strlen(buf);
		}
      
		case LM_SYSTRAY:
		{
			if(bTrayIsReordering)
			{
				return 0;
			}

			NOTIFYICONDATA* nid = (NOTIFYICONDATA*)lParam;
			switch(wParam)
			{
			case NIM_ADD:
				nTrayIconsCount++;
				if (!bAutoReorder) break;
				if(!IsMyNotifWindow(nid->hWnd))
				{
					STrayWndData trayWndData(nid->hWnd, nid->uID);
					trayWndList.insert(trayWndList.begin(), trayWndData);
					if (bFixedTray || bTrayedRoom)
					{
						RemoveSep();
						RecalcIcons();
					}
					ReorderIcons();
				}
				break;
			case NIM_MODIFY:
				if(!IsMyNotifWindow(nid->hWnd))
				{
					STrayWndData trayWndData(nid->hWnd, nid->uID);
					if (!trayWndList.find(trayWndData))
					{
						trayWndList.insert(trayWndList.begin(), trayWndData);
						if (bFixedTray || bTrayedRoom)
						{
							nTrayIconsCount++;
							RemoveSep();
						}
						RecalcIcons();
						ReorderIcons();
					}
				}
				break;
			case NIM_DELETE:
				nTrayIconsCount--;
				STrayWndData trayWndData(nid->hWnd, nid->uID);
				trayWndList.remove(trayWndData);
				if (!bAutoReorder) break;
				RefreshState();
				RecalcIcons();
				break;
			}
		}
		return 0;

	case WM_REORDER_NOTIFY:
		if (bIsIdle) // skip while idle but don't forget to reorder!
		{
			if (!bReorderDelayed)
			{
				bReorderDelayed = true;
				PostMessage(hWndMe, WM_REORDER_NOTIFY, 0, 0);
			}
		}
		else
		{
			bReorderDelayed = false;

			bWaitingForReorder = false;
			bTrayIsReordering = true;
			taskList.move_to_right();
			bTrayIsReordering = false;
		}
		return 0;

	case WM_RECALC_NOTIFY:
		if(bIsIdle) // skip while idle but don't forget to recalc!
		{
			if (!bRecalcDelayed)
			{
				bRecalcDelayed = true;
				PostMessage(hWndMe, WM_RECALC_NOTIFY, 0, 0);
			}
		}
		else
		{
			bRecalcDelayed = false;

			bWaitingForRecalc = false;
			if (bTrayedRoom || bFixedTray)
			{
				UINT nLastOld = nTrayIconsCount;
				nTrayIconsCount = SendMessage(hWndParent, LM_SENDSYSTRAY, 0, (LPARAM) hWndMe);

				if (0 == nTrayIconsCount) // lsapi don't work always? It's seems so...
					nTrayIconsCount = nLastOld;
				
				if (nTrayIconsCount != nLastOld) // to fix ghost icons which comes without NIM_ADD
				{
					RefreshState();
				}
			}
		}
		return 0;

	case LM_RESTORESYSTRAY:
		if(!bModuleInitialized)
		{
			PostMessage(hWndMe, WM_TIMER, INIT_TIMER, 0);
			return 0;
		}
		RecalcIcons();
		RefreshState();
		ReorderIcons();
		return 0;

	case LM_SAVESYSTRAY:
		taskList.safe_clear();
		return 0;

	case LM_REDRAW:
		if (bUseFlashing)
		{
			SetFlashing((HWND)wParam, (BOOL)lParam);
		}
		return 0;

	case WM_TIMER:
		switch(wParam)
		{
		case FLASH_TIMER:
			if (!bModuleEnabled) return 0;
			bIsFlashing = true;
			FlashEmAll();
			bIsFlashing = false;
			break;
		case MAIN_TIMER:
			if (!bModuleEnabled) return 0;
			if (bTrayIsReordering || bIsFlashing) // skip if reordering. will do next time
			{
				if (!bMainTimerDelayed)
				{
					bMainTimerDelayed = true;
					PostMessage(hWndMe, WM_TIMER, MAIN_TIMER, 0);
				}
			}
			else
			{
				bMainTimerDelayed = false;

				bIsIdle = true;
				OnIdle(hWnd);
				bIsIdle = false;
			}
			break;
		case INIT_TIMER:
			KillTimer(hWndMe, INIT_TIMER);
			bModuleInitialized = TRUE;
			if (bModuleEnabled)
			{
				SetModuleOn();
			}
          
			/* TODO later
			if(bFilterProcesses)
			m_win32proc.Init();
			*/
			bModuleInitialized = TRUE;
			break;
		}
		return 0;      
	}
	return DefWindowProc(hWnd, message, wParam, lParam);
}


LRESULT CALLBACK WndProcNotif(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) 
	{
    case WM_ICON_NOTIFY:
        {
          TaskListType::iterator i;
          for(i = taskList.begin(); i != taskList.end(); ++i)
          {
            (*i)->icon.OnTrayNotification(wParam, lParam);
          }
        }
        break;

    case WM_DESTROY:
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}


void RemoveSep()
{
	if(nSepCount > 0 && taskList.size())
	{
		nSepCount--;
		taskList.safe_erase(taskList.begin());
	}
}


void AddSep()
{
  HWND hwndNotif = CreateWindow(szNotifWindowClass, "", WS_CHILD,
  0, 0, 10, 10, hWndMe, NULL, hModule, NULL);
  if(hwndNotif)
  {
    nSepCount++;

    HWND hWndTray=NULL;
    BYTE nCurIconLevel = MAXICONLEVEL;
    HICON hIcon = GetWindowIcon(hwndNotif, nCurIconLevel, 300);
    STaskData *pTask  = new STaskData(hwndNotif, hwndNotif);
    pTask->nIconLevel = nCurIconLevel;
    taskList.insert(taskList.begin(), pTask);
    pTask->icon.Create(hwndNotif, WM_ICON_NOTIFY, "", hIcon, (UINT) hwndNotif);
  }
}


void ReorderIcons()
{
	if (!bWaitingForReorder)
	{
		bWaitingForReorder = true;
		PostMessage(hWndMe, WM_REORDER_NOTIFY, 0, 0);
	}
}


void RecalcIcons()
{
	if (!bTrayedRoom && !bFixedTray) return;
	if (!bWaitingForRecalc)
	{
		bWaitingForRecalc = true;
		PostMessage(hWndMe, WM_RECALC_NOTIFY, 0, 0);
	}
}


void RefreshState()
{
	PostMessage(hWndMe, WM_TIMER, MAIN_TIMER, 0);
}